#ifndef __SPXCONFIG_H__
#define __SPXCONFIG_H__

#define SOPLEX_BUILD_TYPE "None"
#define SOPLEX_VERSION_MAJOR 8
#define SOPLEX_VERSION_MINOR 0
#define SOPLEX_VERSION_PATCH 3
#define SOPLEX_WITH_BOOST
/* #undef SOPLEX_WITH_CPPMPF */
/* #undef SOPLEX_WITH_FLOAT128 */
#define SOPLEX_WITH_GMP
#define SOPLEX_WITH_MPFR
#define SOPLEX_WITH_PAPILO
#define SOPLEX_WITH_ZLIB

#endif
